from typing import Literal
import numpy as np
import numba
import astropy.units as u
import ndfilters

__all__ = [
    "variance_filter",
]


def variance_filter(
    array: np.ndarray | u.Quantity,
    size: int | tuple[int, ...],
    axis: None | int | tuple[int, ...] = None,
    where: bool | np.ndarray = True,
    mode: Literal["mirror", "nearest", "wrap", "truncate"] = "mirror",
) -> np.ndarray:
    """
    Calculate a multidimensional rolling variance.

    Parameters
    ----------
    array
        The input array to be filtered
    size
        The shape of the kernel over which the variance will be calculated.
    axis
        The axes over which to apply the kernel.
        Should either be a scalar or have the same number of items as `size`.
        If :obj:`None` (the default) the kernel spans every axis of the array.
    where
        An optional mask that can be used to exclude parts of the array during
        filtering.
    mode
        The method used to extend the input array beyond its boundaries.
        See :func:`scipy.ndimage.generic_filter` for the definitions.
        Currently, only "mirror", "nearest", "wrap", and "truncate" modes are
        supported.

    Returns
    -------
        A copy of the array with the variance filter applied.

    Examples
    --------

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import scipy.datasets
        import ndfilters

        img = scipy.datasets.ascent()
        img_filtered = ndfilters.variance_filter(img, size=21)

        fig, axs = plt.subplots(ncols=2, sharex=True, sharey=True)
        axs[0].set_title("original image");
        axs[0].imshow(img, cmap="gray");
        axs[1].set_title("filtered image");
        axs[1].imshow(img_filtered, cmap="gray");

    """
    return ndfilters.generic_filter(
        array=array,
        function=_variance,
        size=size,
        axis=axis,
        where=where,
        mode=mode,
    )


@numba.njit
def _variance(
    array: np.ndarray,
    args: tuple[float],
) -> float:
    return np.var(array)



import numba

__all__ = [
    "rectify_index_lower",
    "rectify_index_upper",
]


@numba.njit
def rectify_index_lower(index: int, size: int, mode: str) -> int:
    if mode == "mirror":
        return -index
    elif mode == "nearest":
        return 0
    elif mode == "wrap":
        return index % size
    else:  # pragma: nocover
        raise ValueError


@numba.njit
def rectify_index_upper(index: int, size: int, mode: str) -> int:
    if mode == "mirror":
        return ~(index % size + 1)
    elif mode == "nearest":
        return size - 1
    elif mode == "wrap":
        return index % size
    else:  # pragma: nocover
        raise ValueError



import pytest
import numpy as np
import scipy
import astropy.units as u
import ndfilters


@pytest.mark.parametrize(
    argnames="array",
    argvalues=[
        np.random.random((5, 6, 7)),
        np.random.random((5, 6, 7)) * u.mm,
    ],
)
@pytest.mark.parametrize(
    argnames="kernel,axis",
    argvalues=[
        (np.array([1, 2, 1]), ~0),
        (np.array([1, 2, 3]) / 6, ~0),
        (np.random.random((3,)), ~0),
        (np.random.random((3, 4)), (~1, ~0)),
        (np.random.random((3, 4, 5)), None),
    ],
)
@pytest.mark.parametrize(
    argnames="where",
    argvalues=[
        True,
    ],
)
@pytest.mark.parametrize(
    argnames="mode",
    argvalues=[
        "mirror",
        "nearest",
        "wrap",
        "truncate",
    ],
)
def test_convolve(
    array: np.ndarray | u.Quantity,
    kernel: np.ndarray | u.Quantity,
    axis: None | int | tuple[int, ...],
    where: bool | np.ndarray,
    mode: str,
):

    kwargs = dict(
        array=array,
        kernel=kernel,
        axis=axis,
        where=where,
        mode=mode,
    )

    result = ndfilters.convolve(**kwargs)

    assert result.sum() != 0

    if mode == "truncate":
        return

    axis_ = axis
    if axis_ is None:
        axis_ = np.arange(array.ndim)
    axis_ = np.core.numeric.normalize_axis_tuple(axis_, ndim=array.ndim)

    axis_orthogonal = [ax for ax in range(array.ndim) if ax not in axis_]

    kernel_ = np.expand_dims(kernel, axis=axis_orthogonal)

    result_expected = scipy.ndimage.convolve(
        input=array,
        weights=kernel_,
        mode=mode,
    )

    assert np.allclose(u.Quantity(result).value, result_expected)



"""
Numba-accelerated, n-dimensional filters similar to those in :mod:`scipy.ndimage`.
"""

from ._generic import generic_filter
from ._mean import mean_filter
from ._trimmed_mean import trimmed_mean_filter
from ._variance import variance_filter
from ._convolve import convolve

__all__ = [
    "generic_filter",
    "mean_filter",
    "trimmed_mean_filter",
    "variance_filter",
    "convolve",
]



from typing import Callable, Literal
import pytest
import numpy as np
import numba
import scipy.ndimage
import astropy.units as u
import ndfilters


@numba.njit
def _mean(a: np.ndarray, args: tuple = ()) -> float:
    return np.mean(a)


@pytest.mark.parametrize(
    argnames="array",
    argvalues=[
        np.random.uniform(size=101),
        np.random.uniform(size=101) * u.mm,
    ],
)
@pytest.mark.parametrize(
    argnames="function",
    argvalues=[
        _mean,
    ],
)
@pytest.mark.parametrize(
    argnames="size",
    argvalues=[
        5,
        (5,),
    ],
)
@pytest.mark.parametrize(
    argnames="mode",
    argvalues=[
        "mirror",
        "nearest",
        "wrap",
        "truncate",
        pytest.param("foo", marks=pytest.mark.xfail),
    ],
)
def test_generic_filter(
    array: np.ndarray | u.Quantity,
    function: Callable[[np.ndarray], float],
    size: int | tuple[int, ...],
    mode: Literal["mirror", "nearest", "wrap", "truncate"],
):
    result = ndfilters.generic_filter(
        array=array,
        function=function,
        size=size,
        mode=mode,
    )
    assert result.shape == array.shape
    assert result.sum() != 0

    if mode != "truncate":
        result_expected = scipy.ndimage.generic_filter(
            input=array,
            function=function,
            size=size,
            mode=mode,
        )

        if isinstance(array, u.Quantity):
            assert np.all(result.value == result_expected)
            assert result.unit == array.unit
        else:
            assert np.all(result == result_expected)



from typing import Literal
import pytest
import numpy as np
import scipy.ndimage
import scipy.stats
import astropy.units as u
import ndfilters


@pytest.mark.parametrize(
    argnames="array",
    argvalues=[
        np.random.random(5),
        np.random.random((5, 6)),
        np.random.random((5, 6, 7)) * u.mm,
    ],
)
@pytest.mark.parametrize(
    argnames="size",
    argvalues=[2, (3,), (3, 4), (3, 4, 5)],
)
@pytest.mark.parametrize(
    argnames="axis",
    argvalues=[
        None,
        0,
        -1,
        (0,),
        (-1,),
        (0, 1),
        (-2, -1),
        (0, 1, 2),
        (2, 1, 0),
    ],
)
@pytest.mark.parametrize(
    argnames="mode",
    argvalues=[
        "mirror",
        "nearest",
        "wrap",
    ],
)
def test_mean_filter(
    array: np.ndarray,
    size: int | tuple[int, ...],
    axis: None | int | tuple[int, ...],
    mode: Literal["mirror", "nearest", "wrap", "truncate"],
):
    kwargs = dict(
        array=array,
        size=size,
        axis=axis,
        mode=mode,
    )

    if axis is None:
        axis_normalized = tuple(range(array.ndim))
    else:
        try:
            axis_normalized = np.core.numeric.normalize_axis_tuple(
                axis, ndim=array.ndim
            )
        except np.exceptions.AxisError:
            with pytest.raises(np.exceptions.AxisError):
                ndfilters.mean_filter(**kwargs)
            return

    if isinstance(size, int):
        size_normalized = (size,) * len(axis_normalized)
    else:
        size_normalized = size

    if len(size_normalized) != len(axis_normalized):
        with pytest.raises(ValueError):
            ndfilters.mean_filter(**kwargs)
        return

    result = ndfilters.mean_filter(**kwargs)

    size_scipy = [1] * array.ndim
    for i, ax in enumerate(axis_normalized):
        size_scipy[ax] = size_normalized[i]

    expected = scipy.ndimage.uniform_filter(
        input=array,
        size=size_scipy,
        mode=mode,
    )

    if isinstance(result, u.Quantity):
        assert np.allclose(result.value, expected)
        assert result.unit == array.unit
    else:
        assert np.allclose(result, expected)



from typing import Callable, Literal
import numpy as np
import numba
import astropy.units as u
from ._indices import (
    rectify_index_lower,
    rectify_index_upper,
)

__all__ = [
    "generic_filter",
]


def generic_filter(
    array: np.ndarray | u.Quantity,
    function: Callable[[np.ndarray, tuple], float],
    size: int | tuple[int, ...],
    axis: None | int | tuple[int, ...] = None,
    where: bool | np.ndarray = True,
    mode: Literal["mirror", "nearest", "wrap", "truncate"] = "mirror",
    args: tuple = (),
) -> np.ndarray:
    """
    Filter a multidimensional array using an arbitrary compiled function.

    Parameters
    ----------
    array
        The input array to be filtered
    function
        The function to applied to each kernel footprint.
        This is usually either a Numpy reduction function like :func:`numpy.mean`,
        or a function compiled using :func:`numba.njit`.
        This function must accept a 1D array and a tuple of extra arguments as
        input and return a scalar.
    size
        The shape of the kernel over which the trimmed mean will be calculated.
    axis
        The axes over which to apply the kernel.
        Should either be a scalar or have the same number of items as `size`.
        If :obj:`None` (the default) the kernel spans every axis of the array.
    where
        An optional mask that can be used to exclude parts of the array during
        filtering.
    mode
        The method used to extend the input array beyond its boundaries.
        See :func:`scipy.ndimage.generic_filter` for the definitions.
        Currently, only "mirror", "nearest", "wrap", and "truncate" modes are
        supported.
    args
        Extra arguments to pass to function.

    Examples
    --------

    .. jupyter-execute::

        import numpy as np
        import numba
        import matplotlib.pyplot as plt
        import scipy.datasets
        import ndfilters

        # Download a sample image
        img = scipy.datasets.ascent()

        # Define a compiled function to apply at every
        # kernel footprint.
        @numba.njit
        def function(a: np.ndarray, args: tuple) -> float:
            return np.mean(a)

        # Filter the image using an arbitrary function.
        img_filtered = ndfilters.generic_filter(
            function=function,
            array=img,
            size=21,
        )

        fig, axs = plt.subplots(ncols=2, sharex=True, sharey=True)
        axs[0].set_title("original image");
        axs[0].imshow(img, cmap="gray");
        axs[1].set_title("filtered image");
        axs[1].imshow(img_filtered, cmap="gray");

    """
    if isinstance(array, u.Quantity):
        unit = array.unit
        array = array.value
    else:
        unit = None

    if axis is None:
        axis = tuple(range(array.ndim))
    axis = np.core.numeric.normalize_axis_tuple(axis, ndim=array.ndim)

    if isinstance(size, int):
        size = (size,) * len(axis)
    else:
        if len(size) != len(axis):
            raise ValueError(
                f"{size=} should have the same number of elements as {axis=}."
            )

    axis_numba = ~np.arange(len(axis))[::-1]

    shape = array.shape
    shape_numba = tuple(shape[ax] for ax in axis)

    where = np.broadcast_to(where, shape)

    array_ = np.moveaxis(array, axis, axis_numba)
    where_ = np.moveaxis(where, axis, axis_numba)

    if len(axis) == 1:
        _generic_filter_nd = _generic_filter_1d
    elif len(axis) == 2:
        _generic_filter_nd = _generic_filter_2d
    elif len(axis) == 3:
        _generic_filter_nd = _generic_filter_3d
    else:  # pragma: nocover
        raise ValueError(f"Only 1-3 axes supported, got {axis=}.")

    result = _generic_filter_nd(
        array=array_.reshape(-1, *shape_numba),
        function=function,
        size=size,
        where=where_.reshape(-1, *shape_numba),
        mode=mode,
        args=args,
    )

    result = result.reshape(array_.shape)
    result = np.moveaxis(result, axis_numba, axis)

    if unit is not None:
        result = result << unit

    return result


@numba.njit(parallel=True)
def _generic_filter_1d(
    array: np.ndarray,
    function: Callable[[np.ndarray, tuple], float],
    size: tuple[int],
    where: np.ndarray,
    mode: str,
    args: tuple,
):
    result = np.empty_like(array)

    array_shape_t, array_shape_x = array.shape

    (kernel_shape_x,) = size

    for it in range(array_shape_t):

        for ix in numba.prange(array_shape_x):

            values = np.zeros(shape=size)
            mask = np.zeros(shape=size, dtype=np.bool_)

            for kx in range(kernel_shape_x):

                px = kx - kernel_shape_x // 2
                jx = ix + px

                if jx < 0:
                    if mode == "truncate":
                        continue
                    jx = rectify_index_lower(jx, array_shape_x, mode)
                elif jx >= array_shape_x:
                    if mode == "truncate":
                        continue
                    jx = rectify_index_upper(jx, array_shape_x, mode)

                values[kx] = array[it, jx]
                mask[kx] = where[it, jx]

            if np.any(mask):
                result[it, ix] = function(values[mask], args)
            else:
                result[it, ix] = np.nan

    return result


@numba.njit(parallel=True)
def _generic_filter_2d(
    array: np.ndarray,
    function: Callable[[np.ndarray, tuple], float],
    size: tuple[int, int],
    where: np.ndarray,
    mode: str,
    args: tuple,
):
    result = np.empty_like(array)

    array_shape_t, array_shape_x, array_shape_y = array.shape

    kernel_shape_x, kernel_shape_y = size

    for it in range(array_shape_t):

        for ix in numba.prange(array_shape_x):
            for iy in numba.prange(array_shape_y):

                values = np.zeros(shape=size)
                mask = np.zeros(shape=size, dtype=np.bool_)

                for kx in range(kernel_shape_x):

                    px = kx - kernel_shape_x // 2
                    jx = ix + px

                    if jx < 0:
                        if mode == "truncate":
                            continue
                        jx = rectify_index_lower(jx, array_shape_x, mode)
                    elif jx >= array_shape_x:
                        if mode == "truncate":
                            continue
                        jx = rectify_index_upper(jx, array_shape_x, mode)

                    for ky in range(kernel_shape_y):

                        py = ky - kernel_shape_y // 2
                        jy = iy + py

                        if jy < 0:
                            if mode == "truncate":
                                continue
                            jy = rectify_index_lower(jy, array_shape_y, mode)
                        elif jy >= array_shape_y:
                            if mode == "truncate":
                                continue
                            jy = rectify_index_upper(jy, array_shape_y, mode)

                        values[kx, ky] = array[it, jx, jy]
                        mask[kx, ky] = where[it, jx, jy]

                values = values.reshape(-1)
                mask = mask.reshape(-1)

                if np.any(mask):
                    result[it, ix, iy] = function(values[mask], args)
                else:
                    result[it, ix, iy] = np.nan

    return result


@numba.njit(parallel=True)
def _generic_filter_3d(
    array: np.ndarray,
    function: Callable[[np.ndarray, tuple], float],
    size: tuple[int, int, int],
    where: np.ndarray,
    mode: str,
    args: tuple,
):
    result = np.empty_like(array)

    array_shape_t, array_shape_x, array_shape_y, array_shape_z = array.shape

    kernel_shape_x, kernel_shape_y, kernel_shape_z = size

    for it in range(array_shape_t):

        for ix in numba.prange(array_shape_x):
            for iy in numba.prange(array_shape_y):
                for iz in numba.prange(array_shape_z):

                    values = np.zeros(shape=size)
                    mask = np.zeros(shape=size, dtype=np.bool_)

                    for kx in range(kernel_shape_x):

                        px = kx - kernel_shape_x // 2
                        jx = ix + px

                        if jx < 0:
                            if mode == "truncate":
                                continue
                            jx = rectify_index_lower(jx, array_shape_x, mode)
                        elif jx >= array_shape_x:
                            if mode == "truncate":
                                continue
                            jx = rectify_index_upper(jx, array_shape_x, mode)

                        for ky in range(kernel_shape_y):

                            py = ky - kernel_shape_y // 2
                            jy = iy + py

                            if jy < 0:
                                if mode == "truncate":
                                    continue
                                jy = rectify_index_lower(jy, array_shape_y, mode)
                            elif jy >= array_shape_y:
                                if mode == "truncate":
                                    continue
                                jy = rectify_index_upper(jy, array_shape_y, mode)

                            for kz in range(kernel_shape_z):

                                pz = kz - kernel_shape_z // 2
                                jz = iz + pz

                                if jz < 0:
                                    if mode == "truncate":
                                        continue
                                    jz = rectify_index_lower(jz, array_shape_z, mode)
                                elif jz >= array_shape_z:
                                    if mode == "truncate":
                                        continue
                                    jz = rectify_index_upper(jz, array_shape_z, mode)

                                values[kx, ky, kz] = array[it, jx, jy, jz]
                                mask[kx, ky, kz] = where[it, jx, jy, jz]

                    values = values.reshape(-1)
                    mask = mask.reshape(-1)

                    if np.any(mask):
                        result[it, ix, iy, iz] = function(values[mask], args)
                    else:
                        result[it, ix, iy, iz] = np.nan

    return result



from typing import Literal
import numpy as np
import numba
import astropy.units as u
from ._indices import (
    rectify_index_lower,
    rectify_index_upper,
)

__all__ = [
    "convolve",
]


def convolve(
    array: np.ndarray | u.Quantity,
    kernel: np.ndarray | u.Quantity,
    axis: None | int | tuple[int, ...] = None,
    where: bool | np.ndarray = True,
    mode: Literal["mirror", "nearest", "wrap", "truncate"] = "mirror",
) -> np.ndarray:
    """
    Multidimensional convolution of an array with a given kernel.

    This function differs from :func:`scipy.ndimage.convolve` and
    :func:`astropy.convolution.convolve` because it implements a vectorized
    convolution operation where the kernel is allowed to vary along axes
    orthogonal to the convolution axes.

    Parameters
    ----------
    array
        The input array to be convolved.
    kernel
        The convolution kernel.
        Any non-convolution axes must be broadcastable with `array`.
    axis
        The axes of `array` over which to apply the kernel.
        If :obj:`None`, it is assumed that the convolution is applied to all
        the axes of `array`.
    where
        An optional mask that can be used to exclude elements of `array`
        during the convolution.
    mode
        The method used to extend `array` beyond its boundaries.

    Examples
    --------

    Apply a Gaussian blur to a sample image.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import scipy
        import ndfilters

        # Define arbitrary coordinate system
        # of the kenrel
        x = np.linspace(-1, 1, 51)
        y = np.linspace(-1, 1, 51)
        x, y = np.meshgrid(x, y)

        # Rotate the coordinate system
        t = np.pi / 4
        u = x * np.cos(t) - y * np.sin(t)
        v = x * np.sin(t) + y * np.cos(t)

        # Define the standard deviation
        # in each dimension of the 2D kernel
        sigma_u = 0.5
        sigma_v = 0.1

        # Compute a 2D Gaussian kernel
        kernel_u = np.exp(-np.square(u / sigma_u) / 2)
        kernel_v = np.exp(-np.square(v / sigma_v) / 2)
        kernel = kernel_u * kernel_v
        kernel = kernel / kernel.sum()

        # Download a sample image
        img = scipy.datasets.ascent()

        # Convolve the sample image with the kernel
        img_convolved = ndfilters.convolve(
            array=img,
            kernel=kernel,
        )

        # Plot the results
        fig, axs = plt.subplots(ncols=2, sharex=True, sharey=True)
        axs[0].set_title("original image");
        axs[0].imshow(img, cmap="gray");
        axs[1].set_title("convolved image");
        axs[1].imshow(img_convolved, cmap="gray");
    """
    if isinstance(array, u.Quantity):
        unit = array.unit
        array = array.value
    else:
        unit = None

    if axis is None:
        axis = tuple(range(array.ndim))
    axis = np.array(axis)
    axis = np.core.numeric.normalize_axis_tuple(~axis, ndim=array.ndim)
    axis = ~np.array(axis)

    shape_kernel = list(kernel.shape)
    for ax in axis:
        shape_kernel[ax] = 1

    shape = np.broadcast_shapes(array.shape, shape_kernel, np.shape(where))

    shape_kernel = list(shape)
    for ax in axis:
        shape_kernel[ax] = kernel.shape[ax]

    array = np.broadcast_to(array, shape)
    kernel = np.broadcast_to(kernel, shape_kernel)
    where = np.broadcast_to(where, shape)

    axis_numba = ~np.arange(len(axis))[::-1]
    shape_numba = tuple(shape[ax] for ax in axis)
    shape_kernel_numba = tuple(shape_kernel[ax] for ax in axis)

    array_ = np.moveaxis(array, axis, axis_numba)
    kernel_ = np.moveaxis(kernel, axis, axis_numba)
    where_ = np.moveaxis(where, axis, axis_numba)

    if len(axis) == 1:
        _convolve_nd = _convolve_1d
    elif len(axis) == 2:
        _convolve_nd = _convolve_2d
    elif len(axis) == 3:
        _convolve_nd = _convolve_3d
    else:  # pragma: nocover
        raise ValueError(f"Only 1-3 axes supported, got {axis=}.")

    result = _convolve_nd(
        array=array_.reshape(-1, *shape_numba),
        kernel=kernel_.reshape(-1, *shape_kernel_numba),
        where=where_.reshape(-1, *shape_numba),
        mode=mode,
    )

    result = result.reshape(array_.shape)
    result = np.moveaxis(result, axis_numba, axis)

    if unit is not None:
        result = result << unit

    return result


@numba.njit(parallel=True)
def _convolve_1d(
    array: np.ndarray,
    kernel: np.ndarray,
    where: np.ndarray,
    mode: str,
):
    result = np.zeros_like(array)

    array_shape_t, array_shape_x = array.shape

    _, kernel_shape_x = kernel.shape

    for it in range(array_shape_t):

        for ix in numba.prange(array_shape_x):

            r = 0

            for kx in range(kernel_shape_x):

                px = kx - (kernel_shape_x - 1) // 2
                jx = ix + px

                if jx < 0:
                    if mode == "truncate":
                        continue
                    jx = rectify_index_lower(jx, array_shape_x, mode)
                elif jx >= array_shape_x:
                    if mode == "truncate":
                        continue
                    jx = rectify_index_upper(jx, array_shape_x, mode)

                if where[it, jx]:
                    array_tx = array[it, jx]
                    kernel_tx = kernel[it, ~kx]
                    r += array_tx * kernel_tx

            result[it, ix] = r

    return result


@numba.njit(parallel=True)
def _convolve_2d(
    array: np.ndarray,
    kernel: np.ndarray,
    where: np.ndarray,
    mode: str,
):
    result = np.empty_like(array)

    array_shape_t, array_shape_x, array_shape_y = array.shape

    _, kernel_shape_x, kernel_shape_y = kernel.shape

    for it in range(array_shape_t):

        for ix in numba.prange(array_shape_x):
            for iy in numba.prange(array_shape_y):

                r = 0

                for kx in range(kernel_shape_x):

                    px = kx - (kernel_shape_x - 1) // 2
                    jx = ix + px

                    if jx < 0:
                        if mode == "truncate":
                            continue
                        jx = rectify_index_lower(jx, array_shape_x, mode)
                    elif jx >= array_shape_x:
                        if mode == "truncate":
                            continue
                        jx = rectify_index_upper(jx, array_shape_x, mode)

                    for ky in range(kernel_shape_y):

                        py = ky - (kernel_shape_y - 1) // 2
                        jy = iy + py

                        if jy < 0:
                            if mode == "truncate":
                                continue
                            jy = rectify_index_lower(jy, array_shape_y, mode)
                        elif jy >= array_shape_y:
                            if mode == "truncate":
                                continue
                            jy = rectify_index_upper(jy, array_shape_y, mode)

                        if where[it, jx, jy]:
                            array_txy = array[it, jx, jy]
                            kernel_txy = kernel[it, ~kx, ~ky]
                            r += array_txy * kernel_txy

                result[it, ix, iy] = r

    return result


@numba.njit(parallel=True)
def _convolve_3d(
    array: np.ndarray,
    kernel: np.ndarray,
    where: np.ndarray,
    mode: str,
):
    result = np.empty_like(array)

    array_shape_t, array_shape_x, array_shape_y, array_shape_z = array.shape

    _, kernel_shape_x, kernel_shape_y, kernel_shape_z = kernel.shape

    for it in range(array_shape_t):

        for ix in numba.prange(array_shape_x):
            for iy in numba.prange(array_shape_y):
                for iz in numba.prange(array_shape_z):

                    r = 0

                    for kx in range(kernel_shape_x):

                        px = kx - (kernel_shape_x - 1) // 2
                        jx = ix + px

                        if jx < 0:
                            if mode == "truncate":
                                continue
                            jx = rectify_index_lower(jx, array_shape_x, mode)
                        elif jx >= array_shape_x:
                            if mode == "truncate":
                                continue
                            jx = rectify_index_upper(jx, array_shape_x, mode)

                        for ky in range(kernel_shape_y):

                            py = ky - (kernel_shape_y - 1) // 2
                            jy = iy + py

                            if jy < 0:
                                if mode == "truncate":
                                    continue
                                jy = rectify_index_lower(jy, array_shape_y, mode)
                            elif jy >= array_shape_y:
                                if mode == "truncate":
                                    continue
                                jy = rectify_index_upper(jy, array_shape_y, mode)

                            for kz in range(kernel_shape_z):

                                pz = kz - (kernel_shape_z - 1) // 2
                                jz = iz + pz

                                if jz < 0:
                                    if mode == "truncate":
                                        continue
                                    jz = rectify_index_lower(jz, array_shape_z, mode)
                                elif jz >= array_shape_z:
                                    if mode == "truncate":
                                        continue
                                    jz = rectify_index_upper(jz, array_shape_z, mode)

                                if where[it, jx, jy, jz]:
                                    array_txyz = array[it, jx, jy, jz]
                                    kernel_txyz = kernel[it, ~kx, ~ky, ~kz]
                                    r += array_txyz * kernel_txyz

                    result[it, ix, iy, iz] = r

    return result
